from .context import sgp30

import unittest
class BasicTestSuite(unittest.TestCase):
    """Basic test cases."""
    def test_absolute_truth_and_meaning(self):
        self.assertEqual(sgp30.Crc8().hash([0xBE, 0xEF]), 0x92,"testing doccumentation example")
        self.assertEqual(sgp30.Crc8().hash([1,144]), 76, "First half of default reading")
        self.assertEqual(sgp30.Crc8().hash([0,6]), 39, "second half of default reading, ")




from .context import sgp30

from .mock_smbus2 import MockSMBus
import unittest

class TestConstructor(unittest.TestCase):
    """Basic test cases."""
    def setUp(self):
        self.bus = MockSMBus()
    def test_init(self):
        b=sgp30.sgp30.Sgp30(self.bus)
        b.init_sgp()

    def test_general_call(self):
        b=sgp30.sgp30.Sgp30(self.bus)
        b.i2c_geral_call()
        self.assertEqual(self.bus.addr,0)
        self.assertEqual(self.bus.last,0x06)

class SimpleReadTests(unittest.TestCase):
    """Basic test cases."""
    def setUp(self):
        self.bus = MockSMBus()
        self.sgp=sgp30.sgp30.Sgp30(self.bus)

    def test_read(self):
        self.assertEqual(self.sgp.read_measurements().data,[400,6])

    #No real need to do this super-carefully
    #or i will just be testing my own test code
    #just make sure that CRC test are used, maybe..
    def test_crc_ok(self):
        d=self.sgp.read_measurements()
        self.assertEqual(d.crc_ok,True)

    def test_crc_fail(self):
        bus=MockSMBus(break_crc=True)
        sgp=sgp30.sgp30.Sgp30(bus)
        d=sgp.read_measurements()
        self.assertEqual(d.data,[400,6])
        self.assertEqual(d.crc_ok,False)

class TestBaselineMethods(unittest.TestCase):
    """Basic tests for the load/save baseline-methods"""
    def setUp(self):
        self.bus = MockSMBus()
        self.sgp=sgp30.sgp30.Sgp30(self.bus,baseline_filename="/tmp/sgp-crc-test")
        fbus=MockSMBus(break_crc=True)
        self.fsgp=sgp30.sgp30.Sgp30(fbus)


    def test_save(s):
        s.assertTrue(s.sgp.store_baseline())
        open(s.sgp._baseline_filename).read()

    def test_save_fail(s):
        open(s.sgp._baseline_filename,"w").write("")
        s.assertFalse(s.fsgp.store_baseline(),"ensure broken crc is detected")
        s.assertEquals(open(s.sgp._baseline_filename).read(),"","ensure that broken data is not written")

    def test_save_load(s):
        s.assertTrue(s.sgp.store_baseline())
        s.assertTrue(s.sgp.try_set_baseline())

    def test_load_fail_crc(s):
        open(s.sgp._baseline_filename,"w").write("[133, 152, 0, 138, 32, 0]")
        s.assertFalse(s.sgp.try_set_baseline())

    def test_load_fail_broken_json(s):
        open(s.sgp._baseline_filename,"w").write("[133, 152, 0.. nope")
        s.assertFalse(s.sgp.try_set_baseline())



import smbus2
from smbus2 import SMBusWrapper, SMBus, i2c_msg
from collections import namedtuple 
from functools import partial
from time import sleep, asctime,time
import json
from copy import copy
import requests
import os.path
from .crc import Crc8

DEVICE_BUS = 1
BASELINE_FILENAME = os.path.expanduser("~/.sgp_config_data.txt")

class _cmds():
    """container class for mapping between human readable names and the command values used by the sgp"""
    Sgp30Cmd = namedtuple("Sgp30Cmd",["commands","replylen","waittime"])
    IAQ_INIT=Sgp30Cmd([0x20, 0x03],0,10)
    IAQ_MEASURE=Sgp30Cmd([0x20, 0x08],6,12)
    GET_BASELINE=Sgp30Cmd([0x20, 0x15],6,120)
    SET_BASELINE=Sgp30Cmd([0x20, 0x1e],0,10)
    SET_HUMIDITY=Sgp30Cmd([0x20, 0x61],0,10)
    IAQ_SELFTEST=Sgp30Cmd([0x20, 0x32],3,520)
    GET_FEATURES=Sgp30Cmd([0x20, 0x2f],3,3)
    GET_SERIAL=Sgp30Cmd([0x36, 0x82],9,10)

    @classmethod
    def new_set_baseline(cls,baseline_data):
        cmd = cls.SET_BASELINE
        return cls.Sgp30Cmd(cmd.commands +baseline_data,cmd.replylen,cmd.waittime)

class Sgp30():

    def __init__(self,bus,device_address = 0x58, baseline_filename=BASELINE_FILENAME):
        self._bus = bus
        self._device_addr = device_address
        self._start_time = time()
        self._last_save_time = time()
        self._baseline_filename=baseline_filename

    Sgp30Answer = namedtuple("Sgp30Answer",["data","raw","crc_ok"])
    
    def _raw_validate_crc(s,r):
        a = zip(r[0::3],r[1::3])
        crc = r[2::3] == [Crc8().hash(i) for i in a ]
        return(crc,a)

        return (a,crc_ok)
    def read_write(self,cmd):
        write = i2c_msg.write(self._device_addr,cmd.commands)
        if cmd.replylen <= 0 :
           self._bus.i2c_rdwr(write)
        else:
            read = i2c_msg.read(self._device_addr,cmd.replylen)
            self._bus.i2c_rdwr(write) 
            sleep(cmd.waittime/1000.0)
            self._bus.i2c_rdwr(read)
            r = list(read)
            crc_ok,a=self._raw_validate_crc(r)
            answer = [i<<8 | j for i,j in a]
            return self.Sgp30Answer(answer,r,crc_ok)

    def store_baseline(self):
        with open(self._baseline_filename,"w") as conf:
            baseline=self.read_write(_cmds.GET_BASELINE)
            if baseline.crc_ok == True:
                json.dump(baseline.raw,conf)
                return True
            else:
                #print("Ignoring baseline due to invalid CRC")
                return False

    def try_set_baseline(self):
        try:
            with open(self._baseline_filename,"r") as conf:
                conf = json.load(conf)
        except IOError:
            pass
        except ValueError:
            pass
        else:
            crc,_ = self._raw_validate_crc(conf)
            if len(conf) == 6 and crc == True:
                self.read_write(_cmds.new_set_baseline(conf))
                return True
            else:
                #print("Failed to load baseline, invalid data")
                return False

    def read_measurements(self):
        return self.read_write(_cmds.IAQ_MEASURE)

    def read_selftest(self):
        return self.read_write(_cmds.IAQ_SELFTEST)

    def read_serial(self):
        return self.read_write(_cmds.GET_SERIAL)

    def read_features(self):
        return self.read_write(_cmds.GET_FEATURES)

    def init_sgp(self):
        #print("Initializing SGP30")
        self.read_write(_cmds.IAQ_INIT)

    def i2c_geral_call(self):
        """This attempts to reset _ALL_ devices on the i2c buss
        
        This command issues the i2c-general call RW command that should result
        in all devices aborting any read/write operations and starting to listen
        for new i2c-commands.

        This will usually un-stick the SGP30, but might reset or otherwise
        affect any device on the bus.
        """
        self._bus.write_byte(0,0x06)
        sleep(.1)


def main():
    with SMBusWrapper(1) as bus:
        sgp=Sgp30(bus,baseline_filename=BASELINE_FILENAME+".TESTING")
        print("resetting all i2c devices")
        sgp.i2c_geral_call()
        print(sgp.read_features())
        print(sgp.read_serial())
        sgp.init_sgp()
        print(sgp.read_measurements())
    bus.close()

if __name__ == "__main__":
    main()



class Crc8:
    def __init__(s):
        s.crc=255

    def hash(s,int_list):
        for i in int_list:
            s.addVal(i)
        return s.crc

    def addVal(s,n):
        crc = s.crc
        for bit in range(0,8):
            if ( n ^ crc ) & 0x80:
                crc = ( crc << 1 ) ^ 0x31
            else:
                crc = ( crc << 1 )
            n = n << 1
        s.crc = crc & 0xFF
        return s.crc

#print(Crc8().hash([1,144]))
#print(hex(Crc8().hash([0xBE, 0xEF])))
#[1, 144, 76, 0, 6, 39]



from smbus2 import i2c_msg
from collections import namedtuple 

from .context import sgp30
from time import time

I2CAnswers= namedtuple("I2CAnswers",["answer","min_delay"])
I2A = I2CAnswers
def add_crc(l):
    return l + [sgp30.Crc8().hash(l)]

answers = {
    (0x36, 0x82): I2A(add_crc([0,0]),.4), #GET_SERIAL
    (0x20, 0x15): I2A(add_crc([0,0]), 1), #GET_FEATURES
    (0x20, 0x03): I2A(None, 2), #IAQ_INIT
    (0x20, 0x08): I2A([1, 144, 76, 0, 6, 39], 10), #IAQ_MEASURE
    (0x20, 0x32): I2A(add_crc([0xD4,0x00]), 200), #IAQ_SELFTEST
    (0x20, 0x15): I2A([133, 152, 85, 138, 32, 202], 10), #GET_BASELINE
    (0x20, 0x1e): I2A(None, 10) #SET_BASELINE
}

class MockSMBus:
    def __init__(s,break_crc=False):
        s.status=None
        s.last=None
        s.addr=None
        s._break_crc=break_crc
        s._deadline=time() 

    def _set_deadline(s, t):
        s._deadline=time() + t/1000.

    def i2c_rdwr(s,*msgs):
        for m in msgs:
            if time() < s._deadline:
                raise IOError("To fast buss-access, device not ready")
            if m.flags == 1:
                s._process_read(m)
            else:
                s._process_write(m)

    def write_byte(s,addr,data):
        s.status=None
        s.addr=addr
        s.last=data

    def _process_read(s,msg):
        if s.status == None:
            raise AssertionError("Tried to read before write")
        for i in range(len(s.status)):
            msg.buf[i]=chr(s.status[i])
            if s._break_crc and i%3 == 2:
                msg.buf[i]=chr(s.status[i]^42)
        s.status=None
                

    def _process_write(s,msg):
        a = answers[tuple(msg)[0:2]]
        s._set_deadline(a.min_delay)
        s.status = a.answer
        s.last=msg
    



# -*- coding: utf-8 -*-
# Learn more: https://github.com/zinob/RPI_SGP30

from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()
with open('LICENSE') as f:
    license = f.read()

setup(
    name='sgp30',
    description='Library for reading data from the sensiron SGP30',
    version='0.1.2',
    long_description=readme,
    author='Simon Albinsson',
    author_email='pipmon@zinob.se',
    url='https://github.com/zinob/RPI_SGP30',
    license=license,
    packages=find_packages(exclude=('tests')),
    install_requires=['smbus2']
)
